# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

from typing import Dict,List,NamedTuple,Tuple,Union
import re
import struct
from bisect import bisect
import rtlib

# Join accept RX1 delay - same accross all regions
JaccRxDelay = 5

ChDef = NamedTuple('ChDef', [('freq',int),('minDR',int),('maxDR',int)])

NoneChDef = ChDef(0,0,0)


MAX_EIRPs = [dBm-0.0001 for dBm in [8,10,12,13,14,16,18,20,21,24,26,27,29,30,33,36,1000000]]

def encode_max_eirp(max_eirp:float) -> int:
    return max(0, bisect(MAX_EIRPs, max_eirp)-1)


class DRxDef():
    def __init__(self,dr:int,sf:int,bw:int,dn:int=0) -> None:
        self.dr = dr
        self.sf = sf
        self.bw = bw
        self.dn = dn    # used for down link only
        if self.sf == -1:
            self.name = "DR%d:RFU" % self.dr
        elif self.sf == 0:
            self.name = "DR%d:FSK" % self.dr
        else:
            self.name = "DR%d:SF%dBW%d" % (self.dr, self.sf, self.bw)
    
    @property
    def invalid(self) -> bool:
        return (self.sf == -1)
    
    @property
    def isFSK(self) -> bool:
        return self.sf==0
    
    def __str__(self) -> str:
        return self.name

NONE_DR = DRxDef(0,-1,0)
EU_DRs = [
    DRxDef( 0,12,125),
    DRxDef( 1,11,125),
    DRxDef( 2,10,125),
    DRxDef( 3, 9,125),
    DRxDef( 4, 8,125),
    DRxDef( 5, 7,125),
    DRxDef( 6, 7,250),
    DRxDef( 7, 0,  0), # FSK
]
US_DRs = [
    # UP data rates
    DRxDef( 0,10,125),
    DRxDef( 1, 9,125),
    DRxDef( 2, 8,125),
    DRxDef( 3, 7,125),
    DRxDef( 4, 8,500),
    # DN data rates
    DRxDef( 8,12,500,1),
    DRxDef( 9,11,500,1),
    DRxDef(10,10,500,1),
    DRxDef(11, 9,500,1),
    DRxDef(12, 8,500,1),
    DRxDef(13, 7,500,1),
]
AU_DRs = [
    # UP data rates
    DRxDef( 0,12,125),
    DRxDef( 1,11,125),
    DRxDef( 2,10,125),
    DRxDef( 3, 9,125),
    DRxDef( 4, 8,125),
    DRxDef( 5, 7,125),
    DRxDef( 6, 8,500),
    # DN data rates
    DRxDef( 8,12,500,1),
    DRxDef( 9,11,500,1),
    DRxDef(10,10,500,1),
    DRxDef(11, 9,500,1),
    DRxDef(12, 8,500,1),
    DRxDef(13, 7,500,1),
]


def mkchlist(f:Union[int,float],n:int=1,d:int=200,minDR:int=0,maxDR:int=5,DR:int=-1):
    if DR>=0: minDR=maxDR=DR
    if abs(d) < 2000:
        d = int(d*1e3)
    if f < 10000:
        f = int(f*1e6)
    return [ ChDef(int(f+i*d), minDR, maxDR) for i in range(n) ]


def chmask_from_json(j:List[int]) -> int:
    return j[0] | (j[1]<<64)

def chmask_to_json(chmask:int) -> List[int]:
    return [chmask & ((1<<64)-1), chmask>>64]


class Region(object):
    def __init__(self,
                 regionid:int,
                 name:str,
                 upchannels:List[ChDef],
                 drs:List[DRxDef],
                 num_updr:int,
                 num_rx1droff:int,
                 txpow_steps:int=8,
                 adr_params:Tuple[float,float]=(-9.0,2.75)) -> None:
        self.regionid = regionid
        self.name = name          # region base name
        self.setup = ''
        self.remark = ''
        self.freq_range = [0,0]   # override in subclass
        self.RX2DR = 0
        self.PingDR = 0
        self.RX2Freq = 0   # override in subclass
        self.RxDelay = 1
        self.PingFreq = 0
        self.BCNFreq = []  # type: List[int]
        self.BCNDR = 0
        self.bcn_layout = (0,0,0)  # offset time & info, len -- see loramsg.pack_beacon
        self.upDwellTime = 0
        self.dnDwellTime = 0
        self.dnfreqs = [0]*16
        self.dnfreq_map = {}  # type: Dict[int,int]
        self.maxDR125 = 0  # max default DR (fastest with BW=125kHz)
        self.DRs = [ DRxDef(i,-1,  0) for i in range(16) ]
        for dr in drs:
            self.DRs[dr.dr] = dr
            if not dr.invalid and not dr.dn and dr.bw==125:
                self.maxDR125 = dr.dr
        self.sfbw2dr = { (dr.sf, dr.bw):dr for dr in self.DRs if not dr.dn and dr.sf >= 0 }
        self.txpow = [ -2*i for i in range(txpow_steps) ]
        # map datarate offset to real values - array defined only for valid offsets
        self.effectiveRX1DROffsets = [off for off in range(num_rx1droff)]
        self.num_updr = num_updr
        # snr2dr[dr] => SNR threshold for datarate dr
        # - use dr upto this SNR value
        # - if SNR bigger move to next faster DR
        # - fastest datarate has a biiiigly value snr2dr[num_updr-1]=100.0
        self.snr2dr = [ adr_params[0] + adr_params[1]*i for i in range(num_updr-1) ] + [ 100.0 ]
        self.std_upchannels = len(upchannels)    # default defined
        self.upchannels = upchannels
        self.cflist_type = 0
        self.max_eirp = 16.0    # dBm
        self.duty_cycle = 0.0   # none or e.g. 1/100.0
        self.avail_updr = 0     # type: int
        self.set_chmask()       # all ON + calc avail_updr
        self.max_pdulens   = [250]*16   # default max PDU lengths - without repeater option
        self.max_pdulens_R = [230]*16   #  ditto repeater compatible


    def __str__(self) -> str:
        return self.name+'/'+self.setup if self.setup else self.name

    @property
    def is_intermittent(self) -> bool:
        return self.regionid > 0x10000

    @property
    def base_region(self) -> 'Region':
        return REGIONS[self.name]

    def set_dnfreqs(self, dnfreqs:List[int]) -> None:
        dnlen = min(len(dnfreqs), len(self.upchannels))
        dnfreqs = dnfreqs[:dnlen]
        self.dnfreqs = dnfreqs + [0] * (16-dnlen)   # fewer corner cases
        dnfreq_map = {} # type: Dict[int,int]
        for chidx, dnfreq in enumerate(dnfreqs):
            if not dnfreq: continue
            dnfreq_map[self.upchannels[chidx].freq] = dnfreq
        self.dnfreq_map = dnfreq_map

    def set_chmask(self, chmask:int=0) -> None:
        """Recalculated available DRs (bit mask with 16 bits). Depends on available/defined/enabled channels."""
        if not chmask:
            chmask = self.chmask_allON
        self.chmask = chmask
        minDR = 16
        maxDR = -1
        updr_mask = 0
        for chidx,chdef in enumerate(self.upchannels):
            if (chmask & (1<<chidx)) != 0 and (chdef.minDR < minDR or chdef.maxDR > maxDR):
                updr_mask |= (0xFFFF << chdef.minDR) & (0xFFFF >> (15-chdef.maxDR))
        self.avail_updr = updr_mask

    def is_same(self, other:'Region') -> bool:
        return (self.name        == other.name and         # quick check for base region
                self.upchannels  == other.upchannels and
                self.chmask      == other.chmask and
                self.dnfreqs     == other.dnfreqs and
                self.RX2Freq     == other.RX2Freq and
                self.PingFreq    == other.PingFreq and
                self.max_eirp    == other.max_eirp and
                self.upDwellTime == other.upDwellTime and
                self.dnDwellTime == other.dnDwellTime)

    def check_freq(self, freq:int, exc:bool=True) -> bool:
        if freq < self.freq_range[0] or freq > self.freq_range[1]:
            if not exc: return False
            raise ValueError('Frequency %s not in range: [%s,%s]'
                             % (rtlib.freq_to_str(freq),
                                rtlib.freq_to_str(self.freq_range[0]),
                                rtlib.freq_to_str(self.freq_range[1])))
        return True

    def map_snr2dr(self, measured_snr:float) -> int:
        for thres_idx, thres_snr in enumerate(self.snr2dr):
            if measured_snr < thres_snr:
                break
        return thres_idx

    def fix_updr(self, updr:int, delta:int) -> int:
        """Move given up datarate by delta notches. Delta > 0 towards faster datarates.
        Used in ADR decisions in network server.
        """
        while delta != 0:
            if delta < 0:
                new_updr = max(0,updr+delta)
            else:
                new_updr = min(self.num_updr-1,updr+delta)
            # Never use FSK - LMIC does not implement it
            if (self.avail_updr & (1<<new_updr)) != 0 and not self.DRs[new_updr].isFSK:
                return new_updr
            # DR not available - (reduced) channel plan
            delta += 1 if delta < 0 else -1
        return updr

    def to_dr (self, sf:int, bw:int) -> DRxDef:
        """Convert low level Lora radio parameters into a datarate."""
        if bw >= 1000: bw //= 1000
        return self.sfbw2dr.get((sf,bw),NONE_DR)

    def get_dnfreq(self, freq:int) -> int:
        """Compute down frequency from the up frequency."""
        return self.dnfreq_map.get(freq,freq)
    
    def get_dndr(self, updr:int, rx1droffset:int) -> int:
        """Compute DN data rate from UP data rate and RX1DROffset."""
        return self.fix_updr(updr, -self.effectiveRX1DROffsets[rx1droffset])

    def get_upchannels(self) -> List[ChDef]:
        return self.upchannels
    
    def get_enabled_upchannels(self) -> List[ChDef]:
        chmask = self.chmask
        return [chdef for chidx,chdef in enumerate(self.upchannels) if (chmask & (1<<chidx)) != 0 ]
    
    def get_cflist(self) -> bytes:
        return b''

    def mcmds_linkadr_chmask(self, foptsdn:bytearray, off:int, dr:int, txpow:int=0,
                             chmask:int=0, optneg:int=None) -> int:
        return 0

    @property
    def chmask_allON(self) -> int:
        return (1<<len(self.upchannels))-1



class RegionLikeEU(Region):
    LinkADR_ALLON = 6

    def __init__(self, drs=EU_DRs, num_updr=8, rx2freq:int=0, **kwargs) -> None:
        super().__init__(drs=drs[:num_updr], num_updr=num_updr, **kwargs)
        self.max_pdulens   = [59,59,59,123,250,250,250,250]
        self.max_pdulens_R = [59,59,59,123,230,230,230,230]
        self.RX2Freq  = rx2freq
        self.PingFreq = rx2freq
        self.BCNFreq  = [rx2freq]
        self.RX2DR = 0
        self.PingDR = 0
        self.BCNDR = 3
        self.bcn_layout = (2,8,17)

    def get_cflist(self) -> bytes:
        upchannels = self.upchannels
        nup = len(upchannels)
        stdup = self.std_upchannels
        if stdup == nup:
            return b''
        freq_u3 = [ struct.pack('<I', upchannels[chidx].freq//100)[0:3]
                    for chidx in range(stdup, min(stdup+5, nup)) ]
        # Padding + CFList type 0
        freq_u3.append(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        return (b''.join(freq_u3))[0:16]

    def mcmds_linkadr_chmask(self, foptsdn:bytearray, off:int, dr:int, txpow:int=0,
                             chmask:int=0, optneg:int=None) -> int:
        if not chmask: chmask = self.chmask_allON
        struct.pack_into('<BBHB', foptsdn, off, 3, (dr<<4)|txpow, chmask & 0xFFFF, 0)
        return 1


class Region_EU868(RegionLikeEU):
    def __init__(self) -> None:
        # For test short cuts - EU868 should be #1
        super().__init__(regionid=1, name='EU868',
                         upchannels=mkchlist(f=868.1,n=3,maxDR=5), num_updr=8,
                         num_rx1droff=6, txpow_steps=8,
                         rx2freq=869525000)
        self.max_eirp = 16.0  # dBm
        self.duty_cycle = 1/1000.0   # good for any channel; some channels go higher
        self.freq_range = [863000000,870000000]


class Region_EU433(RegionLikeEU):
    def __init__(self) -> None:
        super().__init__(regionid=3, name='EU433',
                         upchannels=mkchlist(f=433.175,n=3,maxDR=5), num_updr=8,
                         num_rx1droff=7, txpow_steps=8,
                         rx2freq=434665000)
        self.max_eirp = 12.15  # dBm
        self.freq_range = [433175000,434665000]


class RegionLikeUS(Region):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.cflist_type = 1

    def get_cflist(self) -> bytes:
        chmask = self.chmask
        if self.chmask_allON == chmask:
            return b''
        off = 0
        buf = bytearray([0]*15 + [1])    # CFList type 1
        while chmask:
            struct.pack_into('<H', buf, off, chmask & 0xFFFF)
            chmask >>= 16
            off += 2
        return bytes(buf)


def mcmds_linkadr_chmask_likeUS915(reg:Union['Region_US915','Region_AU915'],
                                   foptsdn:bytearray, off:int, dr:int, txpow:int=0,
                                   chmask:int=0, optneg:int=None) -> int:
    if not chmask: chmask = reg.chmask_allON
    if chmask == reg.chmask_allON:
        struct.pack_into('<BBHB', foptsdn, off, 3, (dr<<4)|txpow, (chmask>>64)&0xFF, reg.LinkADR_125ON<<4)
        return 1
    if optneg is not None:
        # LoRa WAN > 1.0 - new ctrl command for blocks of 8+1 channels
        blk8 = 0
        for i in range(8):
            m = (0xFF<<(i*8))|(1<<(64+i))  # channel mask for block i
            c = chmask & m
            if c == m:
                blk8 |= 1<<i
            elif c != 0: # Odd block - can't construct map with 8+1 blocks
                blk8 = 0
                break
        if blk8 > 0:
            # chmaks is a set of blocks
            struct.pack_into('<BBHB', foptsdn, off, 3, (dr<<4)|txpow, blk8, reg.LinkADR_BLK8ON<<4)
            return 1
    # chmask is not a set of blocks but something odd
    # (1) turn off all 125kHz and enable desird 500kHz channels
    # (2) then enable groups of 16x 125kHz channel if at least one channel used
    struct.pack_into('<BBHB', foptsdn, off, 3, (dr<<4)|txpow, (chmask>>64) & 0xFFFF, reg.LinkADR_125OFF<<4)
    ncmds = 1
    for j in range(0, 64, 16):
        mask = (chmask >> j) & 0xFFFF
        if mask:
            struct.pack_into('<BBHB', foptsdn, off+ncmds*5, 3, (dr<<4)|txpow, mask, (j//16)<<4)
            ncmds += 1
    return ncmds


class Region_US915(RegionLikeUS):
    LinkADR_BLK8ON = 5  # v1.1 - enable blocks (block = 8x125kHz + 1x500kHz channels)
    LinkADR_125ON  = 6
    LinkADR_125OFF = 7
    
    def __init__(self):
        # For test short cuts - US903 should be #2
        super().__init__(regionid=2, name='US915',
                         upchannels=mkchlist(f=902.3,n=64,maxDR=3) + mkchlist(f=903.0,n=8,d=1600,DR=4),
                         drs=US_DRs, num_updr=5, num_rx1droff=4, txpow_steps=11, adr_params=(-3.5,2.75))
        self.max_eirp = 30.0
        self.freq_range = [902000000,928000000]
        self.RX2Freq = 923300000
        self.RX2DR  = 8
        self.PingDR = 8
        self.BCNDR  = 8
        self.BCNFreq = [923300000 + chx * 600000 for chx in range(8)]
        self.bcn_layout = (5,11,23)
        self.max_pdulens   = [19,61,133,250,250,None,None,None,61,137,250,250,250,250]
        self.max_pdulens_R = [19,61,133,230,230,None,None,None,41,117,230,230,230,230]

    def mcmds_linkadr_chmask(self, foptsdn:bytearray, off:int, dr:int, txpow:int=0,
                             chmask:int=0, optneg:int=None) -> int:
        return mcmds_linkadr_chmask_likeUS915(self, foptsdn, off, dr, txpow, chmask, optneg)

    def get_dnfreq(self, freq:int) -> int:
        '''Compute down frequency from the up frequency.'''
        assert freq >= 902300000 and freq <= 914900000
        chx = (freq - 902300000) // 100000
        chx = (chx // (2 if (chx&1)==0 else 16)) % 8
        return 923300000 + chx * 600000

    def get_dndr(self, updr:int, rx1droffset:int) -> int:
        '''Compute DN data rate from UP data rate and RX1DROffset.'''
        return max(8, min(updr + 10 - self.effectiveRX1DROffsets[rx1droffset], 13))


class Region_AU915(RegionLikeUS):
    LinkADR_BLK8ON = 5  # enable blocks (block = 8x125kHz + 1x500kHz channels)
    LinkADR_125ON  = 6
    LinkADR_125OFF = 7

    def __init__(self):
        super().__init__(regionid=4, name='AU915',
                         upchannels=mkchlist(f=915.2,n=64,maxDR=5) + mkchlist(f=915.9,n=8,d=1600,DR=6),
                         drs=AU_DRs, num_updr=7, num_rx1droff=6, txpow_steps=11, adr_params=(-9.0,2.75))
        self.max_eirp = 30.0
        self.freq_range = [915000000,928000000]
        self.RX2Freq = 923300000
        self.RX2DR = 8
        self.PingDR = 8
        self.BCNDR = 10
        self.BCNFreq = [923300000 + chx * 600000 for chx in range(8)]
        self.bcn_layout = (3,9,19)
        self.max_pdulens   = [59,59,59,123,250,250,250,None,61,137,250,250,250,250]
        self.max_pdulens_R = [59,59,59,123,230,230,230,None,41,117,230,230,230,230]
    
    def mcmds_linkadr_chmask(self, foptsdn:bytearray, off:int, dr:int, txpow:int=0,
                             chmask:int=0, optneg:int=None) -> int:
        return mcmds_linkadr_chmask_likeUS915(self, foptsdn, off, dr, txpow, chmask, optneg)

    def get_dnfreq(self, freq:int) -> int:
        """Compute down frequency from the up frequency."""
        assert freq >= 915200000 and freq <= 927800000
        chx = (freq - 915200000) // 100000
        chx = (chx // (2 if (chx&1)==0 else 16)) % 8
        return 923300000 + chx * 600000
    
    def get_dndr(self, updr:int, rx1droffset:int) -> int:
        '''Compute DN data rate from UP data rate and RX1DROffset.'''
        return max(8, min(updr + 8 - self.effectiveRX1DROffsets[rx1droffset], 13))


class Region_CN470(RegionLikeUS):
    LinkADR_ALLON = 6

    def __init__(self):
        super().__init__(regionid=5, name='CN470',
                         upchannels=mkchlist(f=470.3,n=96,maxDR=5),
                         drs=EU_DRs[:6], num_updr=6, num_rx1droff=6, txpow_steps=8)
        self.max_eirp = 19.15
        self.freq_range = [470000000,510000000]
        self.RX2Freq = 505300000
        self.RX2DR = 0
        self.PingDR = 0
        self.BCNDR = 2
        self.BCNFreq = [508300000 + chx * 200000 for chx in range(8)]
        self.bcn_layout = (3,9,19)
        self.max_pdulens   = [59,59,59,123,250,250]
        self.max_pdulens_R = [59,59,59,123,230,230]
    
    def mcmds_linkadr_chmask(self, foptsdn:bytearray, off:int, dr:int, txpow:int=0,
                             chmask:int=0, optneg:int=None) -> int:
        if not chmask: chmask = self.chmask_allON
        struct.pack_into('<BBHB', foptsdn, off, 3, (dr<<4)|txpow, 0, Region_CN470.LinkADR_ALLON<<4)
        ncmds = 1
        for i in range(0, 96, 16):
            mask = (chmask >> i) & 0xFFFF
            if mask != 0xFFFF:
                struct.pack_into('<BBHB', foptsdn, off+5*ncmds, 3, (dr<<4)|txpow, mask, (i//16)<<4)
                ncmds += 1
        return ncmds

    def get_dnfreq(self, freq:int) -> int:
        """Compute down frequency from the up frequency."""
        assert freq >= 470200000 and freq < 489400000
        return 500300000 + ((freq - 470200000) // 200000 % 48) * 200000


class Region_CN779(RegionLikeEU):
    def __init__(self) -> None:
        super().__init__(regionid=6, name='CN779',
                         upchannels=mkchlist(f=779.5,n=3,maxDR=5),
                         num_updr=8, num_rx1droff=6, txpow_steps=6,
                         rx2freq=786000000)
        self.max_eirp = 12.15  # dBm
        self.duty_cycle = 1/100.0
        self.freq_range = [779000000,787000000]
        self.BCNFreq = [785000000]  # not the same as RX2Freq!
        self.PingFreq = 785000000


class Region_AS923(RegionLikeEU):
    # These countries cover regionids 30-49 (up to 39 in use)
    COUNTRIES = [
        # country code
        #     regionid
        #           freq_range
        #                   country name
        ('BN', 30, [923000000,925000000], 'Brunei'),
        ('KH', 31, [923000000,925000000], 'Cambodia'),
        ('ID', 32, [923000000,925000000], 'Indonesia'),
        ('JP', 33, [920000000,928000000], 'Japan'),
        ('LA', 34, [923000000,925000000], 'Laos'),
        ('NZ', 35, [915000000,928000000], 'New Zealand'),
        ('SG', 36, [920000000,925000000], 'Singapore'),
        ('TW', 37, [922000000,928000000], 'Taiwan'),
        ('TH', 38, [920000000,925000000], 'Thailand'),
        ('VN', 39, [920000000,925000000], 'Vietnam'),
    ]

    def __init__(self, regionid=7, country:str='', freq_range:List[int]=None, remark='') -> None:
        super().__init__(regionid=regionid, name='AS923'+country,
                         upchannels=mkchlist(f=923.2,n=2,maxDR=5),
                         num_updr=8, num_rx1droff=8, txpow_steps=8)
        self.remark = remark
        self.effectiveRX1DROffsets[-2:] = [-1,-2]
        self.max_eirp = 16.0  # dBm
        self.duty_cycle = 1/1000.0
        if freq_range:
            self.freq_range = freq_range
        if not country:
            self.freq_range = [min(map(lambda v:v[2][0],Region_AS923.COUNTRIES)),
                               max(map(lambda v:v[2][1],Region_AS923.COUNTRIES))]
        self.RX2Freq = 923200000
        self.RX2DR = 2
        self.PingDR = 2
        self.BCNFreq = [923400000]
        self.PingFreq = 923400000
        self.BCNDR = 3
        self.bcn_layout = (2,8,17)
        self.mindr = 0  # if dwell time limit of 400ms then 2
    
    def get_dndr(self, updr:int, rx1droffset:int) -> int:
        '''Compute DN data rate from UP data rate and RX1DROffset.'''
        return min(5, max(self.mindr, updr - self.effectiveRX1DROffsets[rx1droffset]))


class Region_KR920(RegionLikeEU):
    def __init__(self) -> None:
        super().__init__(regionid=8, name='KR920',
                         upchannels=mkchlist(f=922.1,n=3,maxDR=5),
                         num_updr=6, num_rx1droff=6, txpow_steps=8)
        self.max_eirp = 23.0   # dBm - gateway
        self.duty_cycle = 0.0  # none
        self.freq_range = [920900000,923300000]
        self.RX2Freq = 921900000
        self.RX2DR = 0
        self.PingDR = 0
        self.BCNFreq = [923100000]
        self.PingFreq = 923100000
        self.BCNDR = 3
        self.bcn_layout = (2,8,17)


class Region_IN865(RegionLikeEU):
    def __init__(self) -> None:
        super().__init__(regionid=9, name='IN865',
                         upchannels=[ChDef(865062500,0,5),ChDef(865402500,0,5),ChDef(865985000,0,5)],
                         num_updr=6, num_rx1droff=8, txpow_steps=11,
                         rx2freq=866550000)
        self.effectiveRX1DROffsets[-2:] = [-1,-2]
        self.max_eirp = 30.0         # dBm
        self.duty_cycle = 0.0        # good for any channel; some channels go higher
        self.freq_range = [865000000,867000000]
        self.RX2DR = 2
        self.PingDR = 2
        self.BCNDR = 4
        self.bcn_layout = (1,7,19)
    
    def get_dndr(self, updr:int, rx1droffset:int) -> int:
        '''Compute DN data rate from UP data rate and RX1DROffset.'''
        return min(5, max(0, updr - self.effectiveRX1DROffsets[rx1droffset]))


class Region_IL915(RegionLikeEU):
    def __init__(self) -> None:
        super().__init__(regionid=10, name='IL915',
                         upchannels=[ChDef(915700000,0,5),ChDef(915900000,0,5),ChDef(916100000,0,5)],
                         num_updr=8, num_rx1droff=6, txpow_steps=8,
                         rx2freq=916300000)
        self.max_eirp = 14.0         # dBm
        self.duty_cycle = 1/100.0    # good for any channel; some channels go higher
        self.freq_range = [915000000,917000000]
        self.RX2DR = 0
        self.PingDR = 0
        self.BCNDR = 3
        self.bcn_layout = (2,8,17)
    
    def get_dndr(self, updr:int, rx1droffset:int) -> int:
        '''Compute DN data rate from UP data rate and RX1DROffset.'''
        return min(5, max(0, updr - self.effectiveRX1DROffsets[rx1droffset]))


class Region_RU864(RegionLikeEU):
    def __init__(self) -> None:
        super().__init__(regionid=11, name='RU864',
                         upchannels=[ChDef(868900000,0,5),ChDef(869100000,0,5)],
                         num_updr=7, num_rx1droff=8, txpow_steps=8,
                         rx2freq=869100000)
        self.effectiveRX1DROffsets[-2:] = [-1,-2]
        self.max_eirp = 16.0         # dBm
        self.duty_cycle = 0.0        # good for any channel; some channels go higher
        self.freq_range = [864000000,870000000]
        self.PingFreq = 868900000



NoneRegion = Region(regionid=0, name='None',
                    upchannels=[], drs=[],
                    num_updr=0, num_rx1droff=0)

EU868 = Region_EU868()
EU433 = Region_EU433()
US915 = Region_US915()
AU915 = Region_AU915()
CN470 = Region_CN470()
CN779 = Region_CN779()
AS923 = Region_AS923()
KR920 = Region_KR920()
IN865 = Region_IN865()
IL915 = Region_IL915()
RU864 = Region_RU864()

REGIONS = {
    'EU868'  : EU868,
    'EU863'  : EU868,  # alias
    'EU433'  : EU433,
    'US915'  : US915,
    'US902'  : US915,  # alias
    'AU915'  : AU915,
    'CN470'  : CN470,
    'CN779'  : CN779,
    'AS923'  : AS923,
    'KR920'  : KR920,
    'IN865'  : IN865,
    'IL915'  : IL915,
    'RU864'  : RU864,
}

for country, regionid, freq_range, country_name in Region_AS923.COUNTRIES:
    fn = lambda: Region_AS923(regionid, country, freq_range, country_name)
    region = fn()
    globals()['Region_AS923'+country] = fn
    globals()['AS923'+country] = region
    REGIONS[region.name] = region





# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

import struct

from Crypto.Hash import CMAC
from Crypto.Cipher import AES

class crypto:
    @staticmethod
    def encrypt (devkey:bytes, pdu:bytes) -> bytes:
        aes = AES.new(devkey, AES.MODE_ECB)
        return aes.encrypt(pdu)

    @staticmethod
    def decrypt (devkey:bytes, pdu:bytes) -> bytes:
        aes = AES.new(devkey, AES.MODE_ECB)
        return aes.decrypt(pdu)

    @staticmethod
    def calcMicJoin (key:bytes, pdu:bytes) -> int:
        pdu = pdu[:-4]
        cmac = CMAC.new(key, ciphermod=AES)
        cmac.update(pdu)
        mic, = struct.unpack_from('<i', cmac.digest())
        return mic

    @staticmethod
    def calcMic (key:bytes, devaddr:int, seqno:int, dndir:int, pdu:bytes) -> int:
        pdu = pdu[:-4]
        b0 = struct.pack('<BIBiIBB',
                0x49, 0, dndir, devaddr, seqno, 0, len(pdu))
        cmac = CMAC.new(key, ciphermod=AES, msg=b0)
        cmac.update(pdu)
        mic, = struct.unpack_from('<i', cmac.digest())
        return mic

    @staticmethod
    def cipher (key:bytes, devaddr:int, seqno:int, dndir:int, pdu:bytes) -> bytes:
        a0 = struct.pack('<BIBiI', 1, 0, dndir, devaddr, seqno)
        aes = AES.new(key, AES.MODE_CTR, nonce=a0, initial_value=1)
        return aes.encrypt(pdu)

    @staticmethod
    def derive (rootkey:bytes, devnonce:int, appnonce:int, netid:int, keytype:int) -> bytes:
        d = (bytes([keytype])
                + struct.pack('<I', appnonce)[:3]
                + struct.pack('<I', netid)[:3]
                + struct.pack('<H', devnonce)
                + (7 * b'\0'))
        aes = AES.new(rootkey, AES.MODE_ECB)
        return aes.encrypt(d)



# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

from .eui import *
import re

__all__ = ['Eui']

def freq_to_str(f:int) -> str:
    if f==0: return '0'
    s = str(int(f))
    l = len(s)
    m = re.search(r'0*$', s)
    z = len(m.group(0) if m else '')
    g = (l-1)//3
    u = ['Hz','kHz','MHz','GHz'][g]
    a = g*3
    if z < a:
        return s[:l-a]+'.'+s[l-a:l-z]+u
    return s[:l-a]+u



# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

from typing import Any,Mapping,MutableMapping

__all__ = ['Conf', 'Msg', 'State' ]

Conf = Mapping[str,Any]
Msg = MutableMapping[str,Any]
State = MutableMapping[str,Any]



# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

import functools
import re
import struct
from base64 import b16decode as _b16decode


__all__ = ['Eui']


@functools.total_ordering
class Eui(object):
    """Class to represent EUIs. The normal representation is an string with 23 characters
    and a format of HH-...-HH. This matches the JSON and SQL representation.
    There are methods to convert to binary string or 64bit integer.
    """
    
    HEX_REGEX = re.compile("^[0-9A-Fa-f]{16}$")
    """Pure hexadecimal."""
    NUM_REGEX = re.compile("^(\-?\d+|0[xX][0-9A-Fa-f]+)$")
    """Standard decimal or hexadecimal number"""
    REGEX = re.compile("^([0-9a-fA-F]){2}(-([0-9a-fA-F]){2}){7}$")
    """Regular expression matching a well formed EUI"""
    REGEX2 = re.compile("^([0-9a-fA-F]){2}(:([0-9a-fA-F]){2}){7}$")
    """Regular expression matching a well formed EUI using colon instead of dashes"""
    
    
    @staticmethod
    def int2str (num:int) -> str:
        return "-".join(["{0:02X}".format((num>>(i*8)&0xFF)) for i in range(7,-1,-1)])
    
    @staticmethod
    def str2int (s:str) -> int:
        if Eui.HEX_REGEX.match(s):
            return int(s,16)
        if Eui.NUM_REGEX.match(s):
            return int(s,0)
        if Eui.REGEX.match(s):
            b = _b16decode(s.replace("-","").encode('ascii'))
        elif Eui.REGEX2.match(s):
            b = _b16decode(s.replace(":","").encode('ascii'))
        else:
            raise ValueError("Illegal Eui: {}".format(s))
        return struct.unpack_from('>q',b)[0]
    
    @staticmethod
    def from_str(euistr:str) -> 'Eui':
        return Eui(euistr)
    
    @staticmethod
    def from_int(euinum:int) -> 'Eui':
        return Eui(euinum)
    
    @staticmethod
    def from_bytes(euibytes:bytes) -> 'Eui':
        return Eui(euibytes)
    
    def __bool__(self) -> bool:
        # Bool value is not dependent on EUI's int value
        return True
    
    def __index__(self) -> int:
        return struct.unpack_from('>q',self.as_bytes(),0)[0]
    
    def as_int(self) -> int:
        return self.__index__()
    
    def as_bytes(self) -> bytes:
        return _b16decode(self.euistr.replace("-","").encode('ascii'))
    
    def __str__(self):
        return self.euistr
    
    def __repr__(self) -> str:
        return "<{}: {}>".format(self.__class__.__name__, str(self))
    
    def __hash__(self):
        return hash(self.euistr)
    
    def __lt__(self, other):
        return self.euistr < other.euistr
    
    def __eq__(self, other):
        return True if isinstance(other,Eui) and self.euistr==other.euistr else False
    
    def __init__(self, euispec) -> None:
        if isinstance(euispec,Eui):
            self.euistr = euispec.euistr   # type: str
        elif isinstance(euispec,str):
            if Eui.HEX_REGEX.match(euispec):
                self.euistr = Eui.int2str(int(euispec,16))
            elif Eui.NUM_REGEX.match(euispec):
                self.euistr = Eui.int2str(int(euispec,0))
            else:
                if Eui.REGEX.match(euispec):
                    self.euistr = euispec.upper()
                elif Eui.REGEX2.match(euispec):
                    self.euistr = euispec.replace(':','-').upper()
                else:
                    raise ValueError("Illegal EUI: {}".format(euispec))
        elif isinstance(euispec,bytes):
            if len(euispec)!=8:
                raise ValueError("Illegal EUI length: {}".format(len(euispec)))
            self.euistr = "-".join([ "{:02X}".format(euispec[i]) for i in range(8) ])
        elif isinstance(euispec,int):
            self.euistr = Eui.int2str(euispec)
        else:
            raise ValueError("Illegal EUI type: "+str(type(euispec)))





# Copyright (C) 2020-2020 Michael Kuyper. All rights reserved.
# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

from typing import Any,Callable,Dict,List,Tuple,Union
import re

class BF():
    """Bitfield - might also cover entire bytes."""
    def __init__(self, name:str='', bits:Union[int,Tuple[int,int]]=(0,0), signed=False) -> None:
        self.name = name
        self.brange = (bits,bits) if isinstance(bits,int) else bits
        self.signed = signed
        self.value = 0

    def decode(self, v:int) -> int:
        eb = 2<<self.brange[1]
        x = (v & (eb-1)) >>self.brange[0]
        if self.signed and x >= (eb>>1):
            x = x-eb
        self.value = x
        return x

    def encode(self) -> int:
        eb = 2<<self.brange[1]
        lo = self.brange[0]
        return (self.value << lo) & (eb-1)

    def value_as_int(self) -> int:
        return self.value

    def value_as_str(self, fmt:str='s') -> str:
        return '{:d}'.format(self.value_as_int())

    def bits_as_str(self) -> str:
        if self.brange[0] != self.brange[1]:
            return '%d-%d' % self.brange
        return '%d' % self.brange[0]

    def __repr__(self) -> str:
        return '%s:%s=%s' % (self.name, self.bits_as_str(), self.value_as_str('s'))

    def __str__(self) -> str:
        return '%s=%s' % (self.name, self.value_as_str('r'))

    def format(self, fmt:str) -> str:
        if 'S' in fmt:  # structure spec
            return '%s:%s' % (self.name, self.bits_as_str())
        if 'D' in fmt:  # details
            return repr(self)
        return str(self)


class Ux():
    def __init__(self, n, fields:Tuple[BF,...]) -> None:
        self.n = n
        self.fields = fields
        for f in fields:
            self.__setattr__(f.name, f)
        self.value = 0

    def decode(self, ba:bytes, off:int) -> int:
        v = 0
        n = self.n
        for i in range(n):
            v |= ba[off+i]<<(i*8)
        self.value = v
        for f in self.fields:
            f.decode(v)
        return off+n

    def encode(self, ba:bytearray, off:int) -> int:
        v = 0
        for f in self.fields:
            v |= f.encode()
        n = self.n
        for i in range(n):
            ba[off+i] = (v>>(i*8)) & 0xFF
        return off+n

    def __repr__(self) -> str:
        return 'u%d(%s)' % (self.n, ','.join(map(repr, self.fields)))

    def __str__(self) -> str:
        return ' '.join(map(str, filter(lambda a: not isinstance(a,RFU), self.fields)))

    def format(self, fmt:str) -> str:
        """Format element U disables rendering of Ux elements. Element u/text defines the
        separator between Ux elements (default is comma)."""
        sep = ','
        m = re.search(r'[Uu]/([^/]*)/', fmt)
        if m: sep = m.group(1)
        f = sep.join(filter(lambda f:f, map(lambda f:f.format(fmt), self.fields)))
        if 'U' in fmt:
            return f
        return 'u%d(%s)' % (self.n, f)



class U8(Ux):
    def __init__(self, *fields:BF) -> None:
        super().__init__(1, fields)

class U16(Ux):
    def __init__(self, *fields:BF) -> None:
        super().__init__(2, fields)

class U24(Ux):
    def __init__(self, *fields:BF) -> None:
        super().__init__(3, fields)

class U32(Ux):
    def __init__(self, *fields:BF) -> None:
        super().__init__(4, fields)


class BFx(BF):
    def __init__(self, name:str='', bits:Union[int,Tuple[int,int]]=(0,0), signed=False) -> None:
        super().__init__(name,bits,signed)

    def value_as_str(self, fmt:str='s') -> str:
        return 'x{:X}'.format(self.value_as_int())


class RFU(BFx):
    def __init__(self, from_bit=0, to_bit=None, ext:str='') -> None:
        super().__init__('RFU'+ext,bits=(from_bit, to_bit or from_bit))

    def value_as_str(self, fmt:str='s') -> str:
        v = self.value_as_int()
        if v < 10:
            return str(v)
        return 'x{:X}'.format(v)

    def format(self, fmt:str) -> str:
        if 'R' in fmt:
            return ''
        return super().format(fmt)


class Freq(BF):
    def __init__(self, name:str='Freq') -> None:
        super().__init__(name, (0,23))

    def value_as_str(self, fmt:str='s') -> str:
        v = self.value_as_int()
        if v<1000000:
            return '%d' % v
        return '%.1fkHz' % (v/10) if (v % 10) else '%.3fMHz' % (v/1e4)


class Ack(BF):
    def __init__(self, name:str, bit:int) -> None:
        super().__init__(name, bit)

    def value_as_str(self, fmt:str='s') -> str:
        v = self.value_as_int()
        return '1/ACK' if v else '0/NACK'



class Opt():
    CMD = 0
    NAME = ''
    def __init__(self, *args:Ux, **kwargs) -> None:
        self.args = args
        for a in args:
            for f in a.fields:
                self.__setattr__(f.name, f)
        self.set(**kwargs)

    def __repr__(self) -> str:
        a = ' '.join(map(repr, self.args))
        if a: a=' '+a
        return '%s<x%02X%s>' % (self.NAME, self.CMD & 0xFF, a)

    def __str__(self) -> str:
        return '%s<%s>' % (self.NAME, ' '.join(map(str, self.args)))

    def __eq__(self, other:Any) -> bool:
        return (isinstance(other, Opt)
                and self.CMD == other.CMD
                and [a.value for a in self.args] == [a.value for a in other.args])

    def set(self, **kwargs) -> None:
        for k,v in kwargs.items():
            getattr(self,k).value = v

    def format(self, fmt:str) -> str:
        """Format element A print arguments only. a/text/ print command and arguments and
        separates arguments with text (default: space). Char 'C' suppresses the hex command code."""
        sep = ' '
        m = re.search(r'[aA]/([^/]*)/', fmt)
        if m: sep = m.group(1)
        a = sep.join(map(lambda a:a.format(fmt), self.args))
        if 'A' in fmt:
            return a
        chex = 'x%02X ' % (self.CMD & 0xFF) if 'C' not in fmt else ''
        return '%s<%s%s>' % (self.NAME, chex, a)

    def decode(self, ba:bytes, off:int) -> int:
        assert ba[off] == self.CMD & 0xFF
        off += 1
        for a in self.args:
            off = a.decode(ba,off)
        return off

    def encode(self, ba:bytearray, off:int) -> int:
        ba[off] = self.CMD & 0xFF
        off += 1
        for a in self.args:
            off = a.encode(ba,off)
        return off

class OptUp(Opt):
    pass

class OptDn(Opt):
    pass

class Unknown(Opt):
    CMD = 0
    NAME = '?'
    def __init__(self, n:int) -> None:
        self.n = n
        self.un = b''

    def __repr__(self) -> str:
        return '%s<x%s>' % (self.NAME, self.un.hex().upper())

    def __str__(self) -> str:
        return self.__repr__()

    def format(self, fmt:str) -> str:
        return self.__repr__()

    def decode(self, ba:bytes, off:int) -> int:
        n = self.n
        self.un = ba[off:off+n]
        return off+n

    def encode(self, ba:bytearray, off:int) -> int:
        n = self.n
        ba[off:off+n] = self.un
        return off+n


# ================================================================================
# Reset Conf/Ind  --- ABP only ---
# ================================================================================

class ResetInd(OptUp):
    CMD = 1
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(BF('Minor',(0,3)), RFU(4,7)), **kwargs)

class ResetConf(OptDn):
    CMD = 1
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(BF('Minor',(0,3)), RFU(4,7)), **kwargs)


# ================================================================================
# LinkCheck Req/Ans
# ================================================================================

class LinkCheckReq(OptUp):
    CMD = 2

class LinkCheckAns(OptDn):
    CMD = 2
    def __init__(self, **kwargs) -> None:
        super().__init__(
            U8(BF('Margin',(0,7))),
            U8(BF('GwCnt', (0,7))), **kwargs)


# ================================================================================
# LinkADR Req/Ans
# ================================================================================

class LinkADRReq(OptDn):
    CMD = 3
    def __init__(self, **kwargs) -> None:
        super().__init__(
            U8 ( BF ('TXPow',(0,3)), BF('DR',(4,7)) ),
            U16( BFx('ChMask',(0,15)) ),
            U8 ( BF ('NbTrans', (0,3)), BF('ChMaskCntl', (4,6)), RFU(7) ), **kwargs)

class LinkADRAns(OptUp):
    CMD = 3
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('ChAck',0), Ack('DRAck',1), Ack('TXPowAck',2), RFU(3,7)), **kwargs)


# ================================================================================
# DutyCycle Req/Ans
# ================================================================================

class DutyCycleReq(OptDn):
    CMD = 4
    def __init__(self, **kwargs) -> None:
        super().__init__(
            # dc = 2^-MaxDC  (aggr. over all channels)
            U8( BF('MaxDC',(0,3), RFU(4,7)) ), **kwargs)

class DutyCycleAns(OptUp):
    CMD = 4


# ================================================================================
# RXParamSetup Req/Ans
# ================================================================================

class RXParamSetupReq(OptDn):
    CMD = 5
    def __init__(self, **kwargs) -> None:
        super().__init__(
            U8 ( BF('RX2DR',(0,3)), BF('RX1DRoff',(4,6)), RFU(7) ),
            U24( Freq() ), **kwargs)

class RXParamSetupAns(OptUp):
    CMD = 5
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('FreqAck',0), Ack('RX2DRAck',1), Ack('RX1DRoffAck',2), RFU(3,7)), **kwargs)


# ================================================================================
# DevStatus Req/Ans
# ================================================================================

class DevStatusReq(OptDn):
    CMD = 6

class DevStatusAns(OptUp):
    CMD = 6
    def __init__(self, **kwargs) -> None:
        super().__init__(
            U8(BF('Batt',  (0,7))),
            U8(BF('Margin',(0,5), signed=True), RFU(6,7)), **kwargs)


# ================================================================================
# NewChannel Req/Ans
# ================================================================================

class NewChannelReq(OptDn):
    CMD = 7
    def __init__(self, **kwargs) -> None:
        super().__init__(
            U8 ( BF('Chnl',(0,7)) ),
            U24( Freq() ),
            U8 ( BF('MinDR',(0,3)), BF('MaxDR',(4,7)) ), **kwargs)

class NewChannelAns(OptUp):
    CMD = 7
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('ChnlAck',0), Ack('DRAck',1), RFU(2,7)), **kwargs)


# ================================================================================
# RXTimingSetup Req/Ans
# ================================================================================

class RXTimingSetupReq(OptDn):
    CMD = 8
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('Delay',(0,3)), RFU(4,7) ), **kwargs)

class RXTimingSetupAns(OptUp):
    CMD = 8


# ================================================================================
# TXParamSetup Req/Ans
# ================================================================================

class MaxEIRP(BF):
    def __init__(self, name:str='MaxEIRP', bits:Union[int,Tuple[int,int]]=(0,3)) -> None:
        super().__init__(name,bits)
    def value_as_str(self, fmt:str='s') -> str:
        v = self.value_as_int()
        return '%d(%ddBm)' % (v,[8,10,12,13,14,16,18,20,21,24,26,27,29,30,33,36][v])

class TXParamSetupReq(OptDn):
    CMD = 9
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( MaxEIRP(), BF('UpDwell',4), BF('DnDwell',5), RFU(6,7) ), **kwargs)

class TXParamSetupAns(OptUp):
    CMD = 9


# ================================================================================
# DlChannel Req/Ans
# ================================================================================

class DlChannelReq(OptDn):
    CMD = 10
    def __init__(self, **kwargs) -> None:
        super().__init__( U8( BF('Chnl',(0,8)) ), U24(Freq()), **kwargs )

class DlChannelAns(OptUp):
    CMD = 10
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('FreqAck',0), Ack('ChnlAck',1), RFU(2,7)), **kwargs)


# ================================================================================
# Rekey Ind/Conf  --- OTAA only ---
# ================================================================================

class RekeyIndOLD(OptUp):
    """Old TrackNet impl of RekeyInd based on an intermediate spec."""
    CMD = 0x10B   # disambiguate from correct option
    SPECIAL = b'\x0B\x11\x00\x03'
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(BF('Minor',(0,3)), BF('Major',(4,7))),
                          U8(BF('Nonce',(0,7))),
                          U8(BF('DUOFCNT',0), BF('FCNT32',1), BF('RFU',(2,3)), BF('Extra',(4,7))), **kwargs)

class RekeyInd(OptUp):
    CMD = 11
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('Minor',(0,3)), RFU(4,7) ), **kwargs)

class RekeyConf(OptDn):
    CMD = 11
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('Minor',(0,3)), RFU(4,7) ), **kwargs)


# ================================================================================
# ADRParamSetup Req/Ans
# ================================================================================

class ADRParamSetupReq(OptDn):
    CMD = 12
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('Delay',(0,3)), BF('Limit',(4,7)) ), **kwargs)

class ADRParamSetupAns(OptUp):
    CMD = 12


# ================================================================================
# DeviceTime Req/Ans
# ================================================================================

class DeviceTimeReq(OptUp):
    CMD = 13

class DeviceTimeAns(OptDn):
    CMD = 13
    def __init__(self, **kwargs) -> None:
        # time = epoch_secs + frac/2^8
        super().__init__(U32( BF('epoch_secs',(0,31)) ), U8( BF('frac',(0,7)) ), **kwargs)


# ================================================================================
# ForceRejoin Req
# ================================================================================

class ForceRejoinReq(OptDn):
    CMD = 14
    def __init__(self, **kwargs) -> None:
        super().__init__(U16( BF('DR',(0,3)),
                              BF('RejoinType',(4,6)),
                              BF('MaxRetries',(8,10)),
                              BF('Period',(11,13)),
                              RFU(14,15) ), **kwargs)


# ================================================================================
# RejoinParam Req/Ans
# ================================================================================

class RejoinParamReq(OptDn):
    CMD = 15
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('MaxCountN',(0,3)), BF('MaxTimeN',(4,7)) ), **kwargs)

class RejoinParamAns(OptUp):
    CMD = 15
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('TimeAck',0), RFU(1,7)), **kwargs)


# ================================================================================
# PingSlotInfo Req/Ans
# ================================================================================

class Intv(BF):
    def __init__(self, name:str='Intv', bits:Union[int,Tuple[int,int]]=(0,2)) -> None:
        super().__init__(name,bits)
    def value_as_str(self, fmt:str='s') -> str:
        v = self.value_as_int()
        return '%d(%ds)' % (v,1<<v)

class PingSlotInfoReq(OptUp):
    CMD = 16
    def __init__(self, **kwargs) -> None:
        # periodicity: 2^Intv secs
        super().__init__(U8( Intv(bits=(0,2)), RFU(3,7) ), **kwargs)

class PingSlotInfoAns(OptDn):
    CMD = 16


# ================================================================================
# PingSlotChnl Req/Ans
# ================================================================================

class PingSlotChnlReq(OptDn):
    CMD = 17
    def __init__(self, **kwargs) -> None:
        super().__init__(
            U24( Freq() ),
            U8 ( BF('DR',(0,3)), RFU(4,7) ), **kwargs)

class PingSlotChnlAns(OptUp):
    CMD = 17
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('FreqAck',0), Ack('DRAck',1), RFU(2,7)), **kwargs)


# ================================================================================
# BeaconTiming Req/Ans
# ================================================================================

class BeaconTimingReq(OptUp):
    CMD = 18

class BeaconTimingAns(OptDn):
    CMD = 18
    def __init__(self, **kwargs) -> None:
        # deprecated as of 1.1
        super().__init__(U16( BF('Delay',(0,15)) ), U8( BF('Chnl',(0,7)) ), **kwargs)


# ================================================================================
# BeaconFreq Req/Ans
# ================================================================================

class BeaconFreqReq(OptDn):
    CMD = 19
    def __init__(self, **kwargs) -> None:
        super().__init__(U24( Freq() ), **kwargs)

class BeaconFreqAns(OptUp):
    CMD = 19
    def __init__(self, **kwargs) -> None:
        super().__init__( U8(Ack('FreqAck',0), RFU(1,7)), **kwargs)


# ================================================================================
# DeviceMode Ind/Conf
# ================================================================================

class DeviceModeInd(OptUp):
    CMD = 32
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('DevClass',(0,7)) ), **kwargs)

class DeviceModeConf(OptDn):
    CMD = 32
    def __init__(self, **kwargs) -> None:
        super().__init__(U8( BF('DevClass',(0,7)) ), **kwargs)



# ================================================================================
# ----- common lookup tables -----
# ================================================================================

T_CtorOpt = Callable[[],Opt]
T_CtorOptMap = Dict[int,T_CtorOpt]

def _scan_classes(refcls:T_CtorOpt) -> T_CtorOptMap:
    d = {} # type: T_CtorOptMap
    type_refcls = type(refcls)
    for name,cls in globals().items():
        if type(cls) is not type_refcls or cls is refcls or not issubclass(cls,refcls): # type: ignore
            continue
        if not cls.NAME:     # pragma:nobranch
            cls.NAME = name
        d[cls.CMD] = cls
    return d


OPTSUP = _scan_classes(OptUp)
OPTSDN = _scan_classes(OptDn)


def unpack_opts(opts:Union[str,bytes], D:T_CtorOptMap) -> List[Opt]:
    if isinstance(opts,str):
        binopts = bytes.fromhex(opts)
    else:
        binopts = opts
    res = [] # type: List[Opt]
    off = 0
    n = len(binopts)
    while off < n:
        offini = off
        cmd = binopts[off]
        optcls = D.get(cmd)
        if optcls is None:
            opt = Unknown(n-off)  # type: Opt
        else:
            if optcls is RekeyInd and binopts[off:off+4] == RekeyIndOLD.SPECIAL:
                opt = RekeyIndOLD()
            else:
                opt = optcls()
        off = opt.decode(binopts,off)
        res.append(opt)
    return res

def unpack_optsup(opts:Union[str,bytes]) -> List[Opt]:
    return unpack_opts(opts, OPTSUP)

def unpack_optsdn(opts:Union[str,bytes]) -> List[Opt]:
    return unpack_opts(opts, OPTSDN)

def pack_opts(opts:List[Opt]) -> bytes:
    ba = bytearray(256)
    off = 0
    for o in opts:
        off = o.encode(ba,off)
    return bytes(ba[0:off])




import setuptools

setuptools.setup(
        name="liblora",
        version="0.0.3",
        author="Michael Kuyper",
        author_email="mkuyper@mkdata.ch",
        description="A library with functions related to LoRaWAN, derived from Basic MAC",
        url="https://github.com/mkuyper/liblora",
        packages=setuptools.find_packages(),
        install_requires=[ "pycryptodome" ],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            ],
        python_requires='>=3.8',
        )



# Copyright (C) 2020-2020 Michael Kuyper. All rights reserved.
# Copyright (C) 2016-2019 Semtech (International) AG. All rights reserved.
#
# This file is subject to the terms and conditions defined in file 'LICENSE',
# which is part of this source code package.

"""Encoding, decoding, encryption and verification of LoRa messages.
"""

from typing import cast,Optional,Tuple,Union
import math
import struct

from loracrypto import crypto
import rtlib
from rtlib.types import Conf, Msg, State


# Key derivation identifiers
KD_DevAddr    = 240 # proprietary use - derive 32bit DevAddr 
KD_NwkSKeyUp  = 1   # v1.1
KD_NwkSKeyDwn = 3   # spec 1.1 only
KD_NwkSKey    = 1   # spec 1.0 only
KD_AppSKey    = 2   # (de)cipher FRMPayload v1.0+v1.1


class VerifyError(Exception):
    """Exception thrown if verification of message integrity code (MIC) fails on a frame."""
    pass



class MHdr(object):
    FTYPE   = 0xE0
    RFU     = 0x1C
    MAJOR   = 0x03
    DNFLAG  = 0x20  # flags down except for FType.PROP


class Major(object):
    V1 = 0x00


class FrmType(object):
    JREQ = 0x00
    JACC = 0x20
    DAUP = 0x40  # data (unconfirmed) up
    DADN = 0x60  # data (unconfirmed) dn
    DCUP = 0x80  # data confirmed up
    DCDN = 0xA0  # data confirmed dn
    REJN = 0xC0  # rejoin for roaming
    PROP = 0xE0


class FCtrl(object):
    ADREN  =  0x80
    ADRARQ =  0x40
    ACK    =  0x20
    MORE   =  0x10   # in DN direction
    CLASSB =  0x10   # in UP direction
    OPTLEN =  0x0F
    FLAGS  =  0xF0


class DLSettings(object):
    OptNeg   = 0x80    # OptNeg in join accept, RFU in RXParamSetupReq
    RX1DRoff = 0x70
    RX2DR    = 0x0F
    @staticmethod
    def pack(rx1droff:int, rx2dr:int, optneg:bool=True) -> int:
        return (((rx1droff << 4) & DLSettings.RX1DRoff) |
                (rx2dr & DLSettings.RX2DR) |
                (DLSettings.OptNeg if optneg else 0))
    @staticmethod
    def unpack(dlset:int) -> Tuple[int,int,bool]:
        return ((dlset & DLSettings.RX1DRoff) >> 4,
                (dlset & DLSettings.RX2DR),
                (dlset & DLSettings.OptNeg)!=0)

class RXTiming(object):
    RFU     = 0xF0
    RxDelay = 0x0F



def to_hex(b:bytes) -> str:
    return bytes.hex(b).upper()

def to_bytes(s:str) -> bytes:
    return bytes.fromhex(s)

def as_bytes(x:Optional[Union[bytes,str]]) -> Optional[bytes]:
    if isinstance(x,str):
        return bytes.fromhex(x)
    return x


def sanity_check_frame (pdu:bytes) -> int:
    n = len(pdu)
    if n == 0:
        raise ValueError("Received empty frame")
    mhdr = pdu[0]
    if (mhdr & MHdr.MAJOR) != Major.V1:
        raise ValueError("Unsuppored major version: {:02X}".format(mhdr & MHdr.MAJOR))
    if (mhdr & MHdr.RFU) != 0:
        raise ValueError("Unsuppored MHdr.RFU: {:02X}".format(mhdr & MHdr.RFU))
    ftype = mhdr & MHdr.FTYPE
    
    if ftype == FrmType.JREQ or ftype == FrmType.REJN:
        if n != 23:
            raise ValueError("Frame JREQ with illegal length: expecting 23 but len={}".format(n))
        return ftype
    
    if ftype == FrmType.JACC:
        if n != 17 and n != 17+16:
            raise ValueError("Frame JACC with illegal length: expecting 17 or 17+16 but len={}".format(n))
        return ftype
    
    if ftype == FrmType.PROP:
        raise ValueError("Frame PROP detected - no proprietary frame formats supported")
    
    # remaining are DAUP/DADN/DCUP/DCDN - all having the same layout
    if n < 8+4:
        raise ValueError("Data frame too short: min 12 but len={}".format(n))
    foptlen = pdu[5] & FCtrl.OPTLEN
    if n < 8+4+foptlen:
        raise ValueError("Data frame has wrong length: expected 12+{} but len={}".format(foptlen,n))
    if n > 8+4+foptlen:
        port = pdu[8+foptlen]
        if port == 0 and foptlen > 0:
            raise ValueError("Data frame with FPort=0 has FOpts present")
            
    return ftype


def which_fcntdown (pdu:bytes) -> str:
    """Find the frame down counter being referred to by this down dataframe.
    If not a down dataframe return empty string. Assumes a sane frame.
    """
    mhdr = pdu[0]
    ftype = mhdr & MHdr.FTYPE
    if ftype != FrmType.DADN and ftype != FrmType.DCDN:
        return ""    # not a down dataframe
    n = len(pdu)
    foptlen = pdu[5] & FCtrl.OPTLEN
    if n == 8+4+foptlen:
        return 'NFCntDown'   # no port present
    port = pdu[8+foptlen]
    return 'NFCntDown' if port==0 else 'AFCntDown'


def pack_dataframe (mhdr:int = FrmType.DAUP|Major.V1,
                    devaddr:int = 0,
                    fctrl:int = 0,
                    fcnt:int = 0,            # 32bit counter required if nwkskey or appskey present
                    fopts:Optional[Union[bytes,bytearray]] = None,
                    port:Optional[int] = None,
                    payload:Optional[bytes] = None,
                    nwkskey:Optional[bytes] = None,
                    appskey:Optional[bytes] = None) -> bytes:
    """Pack parameters into a data frame pdu with the following layout:
    
    +-----+---------+-----+-------+-------+------+---------+-----+---------+
    |  1  |    4    |  1  |   2   |  0/15 | 0/1  |   0-?   |  4  |   bytes |
    +=====+=========+=====+=======+=======+======+=========+=====+=========+
    | mhdr| devaddr |fctrl|  fcnt | fopts | port | payload | MIC |         |
    +-----+---------+-----+-------+-------+------+---------+-----+---------+
    | >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> | input to MIC  |
    +-----+---------+-----+-------+-------+------+---------+-----+---------+
    |                                            | >>>>>>> | ciphered      |
    +-----+---------+-----+-------+-------+------+---------+-----+---------+
    
    """
    if fopts is None: fopts = b''
    if payload is None: payload = b''
    
    prt = b'' if port is None or port < 0 else struct.pack("B",port)
    
    assert len(prt) == 1 or len(payload) == 0
    assert len(fopts) <= 15
    
    lfopts = len(fopts)
    fctrl = (fctrl & FCtrl.FLAGS) | lfopts
    ftype = mhdr & MHdr.FTYPE
    dndir = 1 if ftype == FrmType.DADN or ftype == FrmType.DCDN else 0
    
    if len(payload) > 0:
        key = nwkskey if port == 0 else appskey
        if key is not None:
            payload = crypto.cipher(key, devaddr, fcnt, dndir, payload)
    
    pdu = struct.pack("<BiBH{}s{}s{}si".format(lfopts, len(prt), len(payload)),
                      mhdr, devaddr, fctrl & 0xFF, fcnt & 0xFFFF, fopts, prt, payload, 0)
    if nwkskey is None:
        return pdu
    mic = crypto.calcMic(nwkskey, devaddr, fcnt, dndir, pdu)
    return pdu[0:-4] + struct.pack('<i', mic)


def peek_dataframe (pdu:bytes) -> Tuple[int,int]:
    """Peek at device short address and 16 bit sequence number of a data frame."""
    addr,fctrl,seqno16 = struct.unpack("<iBH", pdu[1:8])
    return (addr,seqno16)


def unpack_dataframe (pdu:bytes, lastseqno32:int, nwkskey:bytes, appskey:bytes=None) -> Msg:
    """Deconstruct a data frame into individual fields and verify integrity.
    Assumes package is sane (e.g. sanity_check_frame called before).
    If frame not sane then this might fail with an exception other than VerifyError.
    """
    mhdr, addr, fctrl, seqno = struct.unpack("<BiBH", pdu[0:8])
    lower16 = lastseqno32 & 0xFFFF
    upper16 = lastseqno32 >> 16
    if seqno < lower16 and upper16 < 0xFFFF:  # don't roll 32bit over to 0
        upper16 += 1
    seqno = (upper16<<16) + seqno
    ftype = mhdr & MHdr.FTYPE
    dndir = 1 if ftype == FrmType.DADN or ftype == FrmType.DCDN else 0
    pdulen = len(pdu)
    mic = struct.unpack('<i', pdu[pdulen-4:])[0]
    
    if mic != crypto.calcMic(nwkskey, addr, seqno, dndir, pdu):
        raise VerifyError("Verify of data frame failed")
    
    foptlen = fctrl & FCtrl.OPTLEN
    fopts = pdu[8:8+foptlen]         # assuming here 8+4+foptlen <= pdulen - otherwise malformed frame!!!
    if pdulen > 8+4+foptlen:
        port = pdu[8+foptlen]
        payload = pdu[9+foptlen:pdulen-4]
    else:
        port = -1
        payload = b''
    
    # Note: payload to be decrypted by appsrv
    if len(payload) > 0:
        key = nwkskey if port==0 else appskey
        if key is not None:
            payload = crypto.cipher(key, addr, seqno, dndir, payload)
    
    return {'msgtype'   :'updf',
            'MHdr'      : mhdr,
            'DevAddr'   : addr,
            'FCtrl'     : fctrl,
            'FCnt'      : seqno,
            'FOpts'     : fopts,      # field always present, if no options => ''
            'FPort'     : port,
            'FRMPayload': payload,    # field always present, if no payload => ''
            'MIC'       : mic}

def unpack_updf (pdu:bytes) -> Msg:
    """Deconstruct a data frame into individual fields and verify integrity.
    Note, there is no crypto verification nor MIC integrity verification.
    """
    mhdr, addr, fctrl, seqno = struct.unpack("<BiBH", pdu[0:8])
    ftype = mhdr & MHdr.FTYPE
    foptlen = fctrl & FCtrl.OPTLEN
    fopts = pdu[8:8+foptlen] # assuming here 8+4+foptlen <= pdulen - otherwise malformed frame!!!
    pdulen = len(pdu)
    if pdulen > 8+4+foptlen:
        port = pdu[8+foptlen]
        payload = pdu[9+foptlen:pdulen-4]
    else: # we have a malformed frame
        port = -1
        payload = b''
    mic = struct.unpack("<i",pdu[pdulen-4:])[0] # LNS expects a signed 32-bit for the MIC
    return {'msgtype'   :'updf',
            'MHdr'      : mhdr,
            'DevAddr'   : addr,
            'FCtrl'     : fctrl,
            'FCnt'      : seqno,
            'FOpts'     : fopts.hex(),   # field always present, if no options => ''
            'FPort'     : port,
            'FRMPayload': payload.hex(), # field always present, if no payload => ''
            'MIC'       : mic}

# This is special since it considers the two frame counters
# This is only used in device simulation code.
#
def unpack_dndf (sstate:State,  msg:Msg) -> str:
    """Deconstruct a down data frame into individual fields and verify it's integrity.
    Assumes package is sane (e.g. sanity_check_frame called before).
    If frame is not sane then this might fail with an exception other than VerifyError.
    """
    pdu = msg['pdu']
    if isinstance(pdu,str):
        pdu = to_bytes(pdu)
    mhdr, devaddr, fctrl, fcnt = struct.unpack("<BiBH", pdu[0:8])
    foptlen = fctrl & FCtrl.OPTLEN
    fopts = pdu[8:8+foptlen]         # assuming here 8+4+foptlen <= pdulen - otherwise malformed frame!!!
    pdulen = len(pdu)
    if pdulen > 8+4+foptlen:
        port = pdu[8+foptlen]
        payload = pdu[9+foptlen:pdulen-4]
    else:
        port = -1
        payload = b''
    mic = struct.unpack("<i", pdu[pdulen-4:])[0]
    
    # Select the correct down frame counter
    fcnt_key = 'NFCntDown' if port <= 0 else 'AFCntDown'
    s_fcntdn = sstate[fcnt_key]
    lower16 = s_fcntdn & 0xFFFF
    upper16 = s_fcntdn >> 16
    samelow = fcnt == lower16
    if fcnt < lower16 and upper16 < 0xFFFF:  # don't roll 32bit over to 0
        upper16 += 1
    fcnt = (upper16<<16) + fcnt
    ftype = mhdr & MHdr.FTYPE
    assert ftype == FrmType.DADN or ftype == FrmType.DCDN

    nwkskey = to_bytes(sstate['NwkSKeyDwn'])
    while mic != crypto.calcMic(nwkskey, devaddr, fcnt, 1, pdu):
        if not samelow:
            raise VerifyError("Verify of down data frame failed")
        samelow = False    # if lower16 the same try this upper page and next
        fcnt += 0x10000    # if not a replay of last fcnt then maybe jumped ahead
    
    if payload:
        cipkey = nwkskey if port==0 else to_bytes(sstate['AppSKey'])
        payload = crypto.cipher(cipkey, devaddr, fcnt, 1, payload)
        
    msg['MHdr'      ] = mhdr
    msg['DevAddr'   ] = devaddr
    msg['FCtrl'     ] = fctrl
    msg['FCnt'      ] = fcnt
    msg['FOpts'     ] = fopts
    msg['FPort'     ] = port
    msg['FRMPayload'] = payload
    msg['MIC'       ] = mic
    sstate[fcnt_key] = msg['FCnt']
    return fcnt_key


def pack_jreq (nwkkey:bytes,
               joineui:rtlib.Eui=None, deveui:rtlib.Eui=None,
               devnonce:int=0, mhdr:int=Major.V1|FrmType.JREQ) -> bytes:
    """Pack parameters into a JOIN REQUEST pdu with the following layout:
    |  1  |     8   |    8   |    2     |  4  |  bytes - all fields little endian
    | mhdr| joineui | deveui | devnonce | MIC |
    | < - - - - - - - - - - - - - - - ->|           input to MIC
    """
    if joineui is None: joineui = rtlib.Eui("00-00-00-00-00-00-00-00")
    if deveui  is None: deveui  = rtlib.Eui("00-00-00-00-00-00-00-00")
    
    pdu = struct.pack("<BqqHi", mhdr, joineui.as_int(), deveui.as_int(), devnonce & 0xFFFF, 0)
    mic = crypto.calcMicJoin(nwkkey,pdu)
    return pdu[0:-4] + struct.pack("<i", mic)

def unpack_nomic(pdu:bytes) -> Msg:
    """Verify sanity, and deconstruct a frame into individual fields. 
    Note, there is no crypto verification as we do not have keys."""
    #??? why do we have to be JSON serializable?
    ftype = sanity_check_frame(pdu)
    if (ftype==FrmType.JREQ):
        m = unpack_jreq(pdu)
        m['JoinEUI'] = str(m['JoinEUI'])
        m['DevEUI'] = str(m['DevEUI'])
        return m
    m = unpack_updf(pdu)
    return m

def unpack_jreq(pdu:bytes) -> Msg:
    """Unpack JREQ message and extract fields without verifying the integrity of the message..
    This call helps locate the device key and then the integrity must be checked by calling verifyJoinRequest.
    Return None if message is malformed, otherwise return mapping with frame fields.
    """
    mhdr, aeui, deui, devnonce, mic = struct.unpack("<BQQHi", pdu[0:23])
    return { 'msgtype'  : 'jreq', 
             'MHdr'     : mhdr,
             'JoinEUI'  : rtlib.Eui(aeui),
             'DevEUI'   : rtlib.Eui(deui),
             'DevNonce' : devnonce,
             'MIC'      : mic
         }


def verify_jreq(nwkkey:bytes, pdu:bytes) -> None:
    jr = unpack_jreq(pdu)
    if jr['MIC'] != crypto.calcMicJoin(nwkkey,pdu):
        raise VerifyError("Verify of join request failed")


def pack_jacc(nwkkey:Union[bytes,str],
              appnonce:int=0,
              netid:int=0, devaddr:int=0,
              dlset:int=0, rxdly:int=0,
              cflist:Optional[bytes]=None,
              mhdr:int=Major.V1|FrmType.JACC,
              devnonce:Optional[int]=None) -> bytes:
    """Pack parameters into a JOIN ACCEPT pdu with the following layout:
    |  1  |     3    |   3   |    4    |  1  |  1  |  0/16  |  4  |   bytes - all fields little endian
    | mhdr| appnonce | netid | devaddr |dlset|rxdly| cflist | MIC |
    |<- - - - - - - - - - - - - - - - - - - - - - - - - - ->|           input to MIC for v1.0
    |     |<- - - - - - - - - - - - - - - - - - - - - - - - - - ->|     ECB decrypted
    MIC input for v1.1: mhdr|devnonce|appnonce|netid|devaddr|dlset|rxdly|cflist
    """
    optneg = dlset & DLSettings.OptNeg
    nwkkey = cast(bytes, as_bytes(nwkkey))
    pdu = bytearray(17 + (16 if cflist else 0))
    pdu[0] = mhdr
    struct.pack_into("<I",   pdu, 1, appnonce & 0xFFFFFF)
    struct.pack_into("<I",   pdu, 1+3, netid)
    struct.pack_into("<iBB", pdu, 1+3+3, devaddr, dlset, rxdly)
    if cflist:
        struct.pack_into("16s", pdu, 17-4, cflist)
    if optneg:
        micpdu = pdu[0:1] + struct.pack("<H", devnonce) + pdu[1:]
    else:
        micpdu = pdu
    mic = crypto.calcMicJoin(nwkkey, bytes(micpdu))
    struct.pack_into("<i", pdu, len(pdu)-4, mic)
    return bytes(pdu[0:1] + crypto.decrypt(nwkkey, bytes(pdu[1:])))


def unpack_jacc(nwkkey:bytes, pdu:bytes, devnonce:int) -> Msg:
    pdu = bytes(pdu[0:1] + crypto.encrypt(nwkkey, pdu[1:]))
    appnonce = struct.unpack_from("<I", pdu, 1)[0] & 0xFFFFFF
    netid    = struct.unpack_from("<I", pdu, 1+3)[0] & 0xFFFFFF
    devaddr, dlset, rxdly = struct.unpack_from("<iBB", pdu, 1+3+3)
    cflist = None if len(pdu) == 17 else pdu[-20:-4]
    mic = struct.unpack_from("<i", pdu[-4:])[0]
    if dlset & DLSettings.OptNeg:
        micpdu = pdu[0:1] + struct.pack("<H", devnonce) + pdu[1:]
    else:
        micpdu = pdu
    if mic != crypto.calcMicJoin(nwkkey, micpdu):
        raise VerifyError("Verify of join accept failed")
    
    return {
        'MHdr'      : pdu[0],
        'AppNonce'  : appnonce,
        'NetID'     : netid,
        'DevAddr'   : devaddr,
        'DLSettings': dlset,
        'RxDelay'   : rxdly,
        'CFList'    : cflist,
        'MIC'       : mic
    }


def mic_join_request(sstate:State, jrmsg:Msg) -> int:
    """Use the sstate['NwkKey'] to append or verify the MIC of the jrmsg.
    jsmsg must contain the following fields: 'MHdr', 'JoinEUI', 'DevEUI' and 'DevNonce'.
    The method computes a MIC over an encoding of the join request message and
    either sets the field 'MIC' if it is absent or compare the computed value with the value
    of this field. If they don't match a VerifyError is raised.
    """
    pdu = pack_jreq(to_bytes(sstate['NwkKey']),
                    rtlib.Eui(jrmsg['JoinEUI']),
                    rtlib.Eui(jrmsg['DevEUI']),
                    jrmsg['DevNonce'],
                    jrmsg.get('MHdr', Major.V1|FrmType.JREQ))
    mic = struct.unpack("<i", pdu[-4:])[0]
    if 'MIC' in jrmsg:
        if jrmsg['MIC'] != mic:
            raise VerifyError('Verify of join request failed')
    else:
        jrmsg['MIC'] = mic
    return mic


def pack_msg(msg:Msg, mickey:Optional[Union[str,bytes]]=None, cipkey:Optional[Union[str,bytes]]=None) -> bytes:
    """Create a binary encoding of the frame described in msg."""
    mickey = as_bytes(mickey)
    cipkey = as_bytes(cipkey)
    
    fopts = msg.get('FOpts',None)
    fopts = to_bytes(fopts) if fopts else b''
    payld = msg.get('FRMPayload',None)
    payld = to_bytes(payld) if payld else b''
    
    return pack_dataframe(msg['MHdr'],
                          msg['DevAddr'],
                          msg['FCtrl'],
                          msg['FCnt'],
                          fopts,
                          msg.get('FPort',None),
                          payld,
                          cast(bytes,mickey), cast(bytes,cipkey))


def calc_mic(msg:Msg, mickey:Optional[Union[str,bytes]]=None, cipkey:Optional[Union[str,bytes]]=None) -> int:
    """Calculate MIC value from message object."""
    pdu = pack_msg(msg, mickey, cipkey)
    return struct.unpack("<i", pdu[-4:])[0]


def get_mic(pdu:bytes) -> int:
    return struct.unpack("<i", pdu[-4:])[0]


def verify_upmsg(nwkskeyup:Union[str,bytes], devaddr:int, fcnt32:int, pdu:bytes, mic:int) -> bool:
    return mic == crypto.calcMic(cast(bytes,as_bytes(nwkskeyup)), devaddr, fcnt32, 0, pdu)    # 0=dndir (up)


def cipher_msg_nwk(sstate:State, fcntkey:str, msg:Msg) -> None:
    """Cipher the 'FRMPayload' msg field is encrypted or decrypted using
    msg counter field fcntkey (one of 'AFCntDown', 'FCntUp' or 'FCntDown'.
    The msg field 'DevAddr' must hold the device short address.
    """
    assert msg['FPort'] == 0
    payload = msg['FRMPayload']
    if payload:
        try:             devaddr = msg['DevAddr']
        except KeyError: devaddr = sstate['DevAddr']
        if fcntkey == 'FCntUp':
            dndir = 0
            key = 'NwkSKeyUp'
        else:
            dndir = 1
            key = 'NwkSKeyDwn'
        cpdu = crypto.cipher(to_bytes(sstate[key]),
                             devaddr,
                             msg[fcntkey],
                             dndir,
                             to_bytes(payload))
        msg['FRMPayload'] = to_hex(cpdu)


def cipher_msg_app(sstate:State, fcntkey:str, msg:Msg) -> None:
    """Cipher the 'FRMPayload' msg field is encrypted or decrypted using
    msg counter field fcntkey (one of 'AFCntDown', 'FCntUp' or 'FCntDown'.
    The msg field 'DevAddr' must hold the device short address.
    """
    payload = msg['FRMPayload']
    if payload:
        try:             devaddr = msg['DevAddr']
        except KeyError: devaddr = sstate['DevAddr']
        cpdu = crypto.cipher(to_bytes(sstate['AppSKey']),
                             devaddr,
                             msg[fcntkey],
                             0 if fcntkey == 'FCntUp' else 1,
                             to_bytes(payload))
        msg['FRMPayload'] = to_hex(cpdu)


def recipher_msg_app(sstate:State, msg:Msg, prevaddr:Optional[int]=None, prevskey:str='') -> None:
    """Decipher the 'FRMPayload' msg field and encipher it again using
    the current session key and frame counter."""
    payload   = msg['FRMPayload']
    devaddr   = sstate['DevAddr']
    afcntdown = sstate['AFCntDown']
    appskey   = to_bytes(sstate['AppSKey'])
    if prevaddr is None:
        pdu = crypto.cipher(appskey, devaddr, msg['AFCntDown'], 1, to_bytes(payload))
    else:
        pdu = crypto.cipher(to_bytes(prevskey), prevaddr, msg['AFCntDown'], 1, to_bytes(payload))
    pdu = crypto.cipher(appskey, devaddr, afcntdown, 1, pdu)
    msg['AFCntDown'] = afcntdown
    msg['FRMPayload'] = to_hex(pdu)


def mic_msg(sstate:State, msg:Msg) -> None:
    """Compute message integrity code (MIC) over the fields contained in msg.
    This method encodes the fields and computes the MIC with the key sstate['NwkSKey{Up,Dwn}'] 
    depending on indicated direction in field msg['MHdr'].
    Msg must contain the following fields: 'MHdr', 'DevAddr', 'FCtrl', 'FCnt', 'FOpts',
    and optional fields 'FPort', 'FRMPayload'.
    This method verify the computed MIC with msg['MIC'] if this field is present and
    will raise an VerifyError if not equal. If the field msg['MIC'] is not present
    it will add the computed MIC to msg.
    """
    mhdr = msg['MHdr']
    ftype = mhdr & MHdr.FTYPE
    key = 'NwkSKeyDwn' if ftype == FrmType.DADN or ftype == FrmType.DCDN else 'NwkSKeyUp'
    code = pack_dataframe(mhdr,
                          msg['DevAddr'],
                          msg['FCtrl'],
                          msg['FCnt'],       # Note: must be 32bit frame counter
                          to_bytes(msg.get('FOpts','')),
                          msg.get('FPort',None),
                          to_bytes(msg.get('FRMPayload','')),
                          to_bytes(sstate[key]))
    mic = struct.unpack("<i", code[-4:])[0]
    if 'MIC' in msg:
        if msg['MIC'] != mic:
            raise VerifyError('Verify of data frame failed')
    else:
        msg['MIC'] = mic
    return mic


# ================================================================================
#
# DERIVE
#
# ================================================================================


def derive_apps_keys (sstate:State, msg:Msg, netid:int, lwversion10:bool=False) -> None:
    devnonce = msg['DevNonce']
    appnonce = msg['AppNonce']
    appkey   = sstate['NwkKey' if lwversion10 else 'AppKey']
    appskey = crypto.derive(to_bytes(appkey), devnonce, appnonce, netid, KD_AppSKey)
    sstate['AppSKey' ] = to_hex(appskey)


def derive_nwks_keys (sstate:State, sess:State, netid:int, lwversion10:bool=False) -> None:
    devnonce = sess['DevNonce']
    appnonce = sess['AppNonce']
    nwkkey   = sstate['NwkKey']

    # Derivation for V1.1/0 devices
    nwkskeyup  = to_hex(crypto.derive(to_bytes(nwkkey), devnonce, appnonce, netid, KD_NwkSKeyUp))
    nwkskeydwn = to_hex(crypto.derive(to_bytes(nwkkey), devnonce, appnonce, netid, KD_NwkSKeyDwn))
    sess['NwkSKeyUp' ] = nwkskeyup
    sess['NwkSKeyDwn'] = nwkskeyup if lwversion10 else nwkskeydwn
    if 'AppKey' in sstate:
        appkey = sstate['NwkKey' if lwversion10 else 'AppKey']
        appskey = crypto.derive(to_bytes(appkey), devnonce, appnonce, netid, KD_AppSKey)
        sess['AppSKey' ] = to_hex(appskey)


# ================================================================================
#
# AIRTIME
#
# ================================================================================

def airtime_symbols (bw:int=125, sf:int=7, nsyms:int=1) -> float:
    """Calculate airtime of n symbols in microseconds."""
    # Symbol rate / time for one symbol (secs)
    if bw in [125,250,500]:
        bw *= 1000
    assert bw in [125000,250000,500000]
    assert sf >= 7 and sf <= 12
    Rs = bw / (1<<sf)
    Ts = 1e6/Rs
    # Length/time of preamble
    return math.ceil( nsyms*Ts )


def airtime (bw:int=125, sf:int=7, plen:int=0, cr:int=1, ih:int=0, crc:int=1, dro:Optional[int]=None, npreamble:int=8) -> float:
    """Calculate air time of Lora packet in microseconds.
    The parameters are:
    | bandwidth (bw) one of 125,250,500 (kHz) or 12500,25000,50000 (Hz)
    | spreading factor (sf): one 7,..,12 or 0 for FSK @ 50kBit/s
    | packet payload length (plen): 0..255
    | coding rate (cr): 1..4
    | implicit header: 0 (explicit) or 1 (implicit)
    | cyclic redundancy check (crc): 0 (absent), 1 (present)
    | datarate optimization (dro): None sf=7,..,10: 0, sf=11,12 1, 0 (off), 1 (on)
    | number of preamble symbols (npreamble): default 8
    Return number of microseconds.
    """
    if sf == 0:
        # rough estimate of FSK @ 50kBit/s
        extra = 5+3+1+2   # preamble, syncword, len, crc
        bits = (plen+extra) * 8
        return  math.ceil(bits * 1e6 / 50e3)    # bits * us/sec / kBit/s
    
    if bw in [125,250,500]:
        bw *= 1000
    # dro=None 7-10 off, 11,12 on
    if dro is None:
        dro = int((sf>=11 and bw==125000) or (sf==12 and bw==250000))
    
    assert bw in [125000,250000,500000]
    assert sf >= 7 and sf <= 12
    assert plen >= 0 and plen <= 255
    assert cr >= 1 and cr <= 4
    assert ih==0 or ih==1
    assert crc==0 or crc==1
    assert dro==0 or dro==1
    
    # Symbol rate / time for one symbol (secs)
    Rs = bw / (1<<sf)
    Ts = 1/Rs
    # Length/time of preamble
    Tpreamble = (npreamble + 4.25) * Ts
    # Symbol length of payload and time
    tmp = math.ceil((8*plen - 4*sf + 28 + 16*crc - ih*20) / (4*sf-dro*8)) * (cr+4)
    npayload = 8 + max(0,tmp)
    Tpayload = npayload * Ts
    # Time on air 
    Tonair = Tpreamble + Tpayload
    # return us secs
    return math.ceil(Tonair*1e6);


